/**
 * 
 * miniwin.c:
 * author: david mueller, 2006
 * e-mail: d.mueller@e-aktuell.at OR davidlukas.mueller@reflex.at
 * version: 0.1.0 alpha
 * released: 04/30/2006 20:30 UTC
 * feel free to reuse anything you may find interesting.
 *
**/

/*

Last Changes

- code fixes
- added animation on mindowing/demindowing
- prepared code for even more animation / eye candy

Changes to 0.0.13

- fixed many, _many_ bugs
- added (once again) the dock window
- added dock window dynamic background (pretty useless and error-prone, but it shows you the possibilities)
- - changes background on click on a window
- improved default settings
- - please set demindow_key to Button3, otherwise you will not be able to move your mindows around
- - - or change your default move key bindings.

Changes to 0.0.12(b,c)

ATTENTION: please run a gconftool-2 --recursive-unset /apps/compiz/plugins/miniwin as there are many
           options which have changed their name and/or were removed
           
- code rewrite: prepared for 0.1.0
- - introduced event handling for mindows
- - separated the taskbar code from the basic mindow code
- fixed: mindows as well as windows remember their position (independent from each other)
- fixed: included QuinnStorm's bugfixes
- changed taskbar behaviour
- - mindow_key doesn't put mindows in the taskbar, only minimizing them does
- - that means: it is now possible to have (different) mindows on your desktop _and_ in the taskbar
- added option only_move_up (thanks to QuinnStorm)
- - good for a bottom aligned taskbar (mindows only grow up and to both sides instead of to all sides)
- removed option bring_back_centered
- removed the dock window (temporarily) as it seems to cause *many* problems

known bugs:
- if you hover mindows very fast and often, the may move around

Changes to 0.0.11

- fixed (again, hopefully) the taskbar decoration bug
- improved scaling behaviour
- - mindows no scale about their center points instead of the top left hand corner (thanks to GeorgeNorton)
- improved taskbar facilities
- added option show_dock (only affects taskbar-mode)
- begin: changed the code layout in order to have moveable mindows

Changes to 0.0.10

- fixed some window refreshing issues and the 100% CPU loadage thing (thanks to QuinnStorm)
- fixed: the taskbar window is now only created when use_as_taskbar is activated
- fixed some minor taskbar things
- fixed of course some minor bugs
- 
HideWindow event received!
fixed (hopefully) the taskbar window decoration problem
- fixed (hopefully): the firefox redraw problem
- improved taskbar behaviour
- - click on taskbar hides it and all the mindows on it
- - taskbar is only shown when there are mindows
- added option taskbar_hide_on_click

known issues:
- flickering on mindowing and demindowing (will be fixed once the smooth scale option will appear)
- taskbar appears in gnome window list as "unnamed window"
- taskbar moving/resizing is not (yet) supported
- - ATTENTION: it is also reported that moving the taskbar may crash compiz
- the taskbar currently is placed on the bottom of the screen, so bad luck for those who have their panel their
- - (will be changed in later versions)

Changes to 0.0.9

- fixed the 100% CPU problem (in parts)
- - the screen now only is painted with the transformed mask, when mindows are hovered
- fixed some taskbar issues
- fixed even more bugs
- added a taskbar window
- added option taskbar_opacity

Changes to 0.0.8

- fixed: switcher now can deiconify mindows
- - problem: as I fixed the threading thing, mindows now also get listed in switcher!
- fixed: the threading issue
- - threading no is NOT anymore necessary (what a work!)
- fixed some more or less important things
- fixed: mindows now preserve their window's aspect ratio (thanks to VRWarper)

Changes to 0.0.7

- fixed the bug that caused white windows when iconify_on_minimize is set (hopefully!)
- fixed the keyboard input bug, which caused the focused window to loose its keyboard focus on a mindow mouseover
- fixed: destroying a window while mini'd will cause the mindow to disappear as well
- fixed: mindows now only scale on hover, when scale_on_hover_size is NOT set to 1.0
- fixed AIGLX code (thanks to GeorgeNorton)
- fixed: mindow and window don't appear no more longe at the same time
- improved taskbar facilities
- - proper arrangement of mindows in the taskbar
- - taskbar may have more than one row

known bugs:
- moving mindows is dangerous - windows will loose their keyboard focus

Changes to 0.0.6

- fixed problem with windows whose height is too long
- - miniwin scaled only per width, but now scales the height too, when a mindow would get too long
- fixed aiglx upside down windows problem (thanks to GeorgeNorton)
- - added option runs_on_aiglx
- fixed some problems with switcher
- improved the default settings (thanks to g14)
- improved taskbar facilities
- added option iconify_on_minimize (yet deactivated by default)
- added option on_all_workspaces (yet deactivated by default)
- removed plugin's dependencies as they were only for testing

and there were of course some minor bugfixes...

Changes to 0.0.5b

- fixed the border problem (mindows now never will have borders, thanks to QuinnStorm)
- changed miniwinBindPixmap -- hopefully the crop-effect now has gone (another thanks to QuinnStorm :)
- changed the bring back behaviour:
- - windows now won't reappear where the mindow was when closed
- - windows now reappear where the went away (default)
- - added option bring_back_centered
- - - if checked, windows will reappear in the middle of the screen -- TODO: test it (fullscreen? maximized?)
- added some extra logging (debug) output
- use miniwin as a taskbar -- TODO: there are "bugs" in the placing routines. need another concept.
- - customizable spacing between mindows in the "taskbar" (taskbar_spacing)

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h> //don't blame me, blame the weatherman!
#include <unistd.h>

#include <X11/Xatom.h>
#include <X11/extensions/Xrender.h>

#include <compiz.h>


static char *winType[] = {
    "Splash",
    "Normal",
    "Dialog",
    "ModalDialog",
};
#define N_WIN_TYPE (sizeof (winType) / sizeof (winType[0]))

static int displayPrivateIndex;


#define GET_MINIWIN_DISPLAY(d) \
         ((MiniwinDisplay *) (d)->privates[displayPrivateIndex].ptr)
         
#define MINIWIN_DISPLAY(d) \
         MiniwinDisplay* md = GET_MINIWIN_DISPLAY(d)
         
#define GET_MINIWIN_SCREEN(s, fd) \
   ((MiniwinScreen*) (s)->privates[(fd)->screenPrivateIndex].ptr)
   
#define MINIWIN_SCREEN(s) \
   MiniwinScreen* ms = GET_MINIWIN_SCREEN(s, GET_MINIWIN_DISPLAY(s->display))
   
#define GET_MINIWIN_WINDOW(w, ss)					  \
    ((MiniwinWindow *) (w)->privates[(ss)->windowPrivateIndex].ptr)

#define MINIWIN_WINDOW(w)					       \
    MiniwinWindow *mw = GET_MINIWIN_WINDOW  (w,		       \
		      GET_MINIWIN_SCREEN  (w->screen,	       \
		      GET_MINIWIN_DISPLAY (w->screen->display)))

#ifndef _NDEBUG
#define DEBUG(X) fprintf (stderr, X);
#else
#define DEBUG(X) 
#endif

#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))

#define MINIWIN_MINDOW_WIDTH_DEFAULT 80
#define MINIWIN_MINDOW_WIDTH_MIN 20
#define MINIWIN_MINDOW_WIDTH_MAX 512

#define MINIWIN_SCALE_ON_HOVER_DEFAULT 1.1f
#define MINIWIN_SCALE_ON_HOVER_MIN 0.90f
#define MINIWIN_SCALE_ON_HOVER_MAX 10.0f
#define MINIWIN_SCALE_ON_HOVER_STEP 0.01f

#define MINIWIN_OPACITY_ON_HOVER_DEFAULT 90
#define MINIWIN_OPACITY_ON_HOVER_MIN 0
#define MINIWIN_OPACITY_ON_HOVER_MAX 100

#define MINIWIN_MINDOW_OPACITY_DEFAULT 60
#define MINIWIN_MINDOW_OPACITY_MIN 0
#define MINIWIN_MINDOW_OPACITY_MAX 100

#define MINIWIN_DOCK_OPACITY_DEFAULT 40
#define MINIWIN_DOCK_OPACITY_MIN 0
#define MINIWIN_DOCK_OPACITY_MAX 100

#define MINIWIN_ANIMATION_TIME_DEFAULT 500
#define MINIWIN_ANIMATION_TIME_MIN 0
#define MINIWIN_ANIMATION_TIME_MAX 2000

#define MINIWIN_MINDOWS_ALWAYS_ON_TOP_DEFAULT TRUE
#define MINIWIN_USE_AS_TASKBAR_DEFAULT TRUE
#define MINIWIN_MINDOWS_ON_ALL_WORKSPACES_DEFAULT TRUE
#define MINIWIN_DOCK_HIDE_ON_CLICK_DEFAULT FALSE
#define MINIWIN_RUNS_ON_AIGLX_DEFAULT FALSE
#define MINIWIN_SHOW_DOCK_DEFAULT TRUE
#define MINIWIN_ONLY_MOVE_UP_DEFAULT FALSE

#define MINIWIN_TASKBAR_SPACING_DEFAULT 50
#define MINIWIN_TASKBAR_SPACING_MIN -MINIWIN_MINDOW_WIDTH_MIN
#define MINIWIN_TASKBAR_SPACING_MAX MINIWIN_MINDOW_WIDTH_MAX

#define MINIWIN_MINDOW_BUTTON_DEFAULT 	Button1
#define MINIWIN_MINDOW_MODIFIERS_DEFAULT	(ControlMask | ShiftMask)

#define MINIWIN_DEMINDOW_BUTTON_DEFAULT 	Button3
#define MINIWIN_DEMINDOW_MODIFIERS_DEFAULT 0

#define MINIWIN_DOCK_SET_BACKGROUND_BUTTON_DEFAULT Button3
#define MINIWIN_DOCK_SET_BACKGROUND_MODIFIERS_DEFAULT (ControlMask | ShiftMask)

enum
{
   //General mindow settings
   MINIWIN_SCREEN_OPTION_MINDOW_WIDTH,
   MINIWIN_SCREEN_OPTION_ANIMATION_TIME,
   MINIWIN_SCREEN_OPTION_MINDOW_OPACITY,
   MINIWIN_SCREEN_OPTION_SCALE_ON_HOVER,
   MINIWIN_SCREEN_OPTION_OPACITY_ON_HOVER,
   MINIWIN_SCREEN_OPTION_MINDOWS_ALWAYS_ON_TOP,
   MINIWIN_SCREEN_OPTION_MINDOWS_ON_ALL_WORKSPACES,
   //Iconify settings
   MINIWIN_SCREEN_OPTION_MINDOW_KEY,
   MINIWIN_SCREEN_OPTION_DEMINDOW_KEY,
   //Taskbar settings
   MINIWIN_SCREEN_OPTION_USE_AS_TASKBAR,
   MINIWIN_SCREEN_OPTION_TASKBAR_SPACING,
   MINIWIN_SCREEN_OPTION_SHOW_DOCK,
   MINIWIN_SCREEN_OPTION_DOCK_OPACITY,
   MINIWIN_SCREEN_OPTION_DOCK_HIDE_ON_CLICK,
   MINIWIN_SCREEN_OPTION_ONLY_MOVE_UP,
   MINIWIN_SCREEN_OPTION_DOCK_SET_BACKGROUND_KEY,
   //Internal settings
   MINIWIN_SCREEN_OPTION_WINDOW_TYPE,
   MINIWIN_SCREEN_OPTION_RUNS_ON_AIGLX,
   MINIWIN_SCREEN_OPTION_NUM
};

typedef enum _MiniwinWindowState
{
   //Normal window states
   MINIWIN_WINDOW_STATE_NORMAL_NONE,
   MINIWIN_WINDOW_STATE_NORMAL_INITIALIZING,
   MINIWIN_WINDOW_STATE_NORMAL_HIDING,
   MINIWIN_WINDOW_STATE_NORMAL_UNHIDING,
   MINIWIN_WINDOW_STATE_NORMAL_HIDDEN,
   MINIWIN_WINDOW_STATE_NORMAL_WAITING,
   //Dock window states
   MINIWIN_WINDOW_STATE_DOCK_CREATED,
   MINIWIN_WINDOW_STATE_DOCK_NONE,
   MINIWIN_WINDOW_STATE_DOCK_HIDDEN,
   //Mindow states
   MINIWIN_WINDOW_STATE_MINDOW_CREATED,
   MINIWIN_WINDOW_STATE_MINDOW_HIDDEN,
   MINIWIN_WINDOW_STATE_MINDOW_ICON,
   MINIWIN_WINDOW_STATE_MINDOW_TASKBAR
} MiniwinWindowState;

#define MwmHintsDecorations (1L << 1)

typedef struct {
    unsigned long flags;
    unsigned long functions;
    unsigned long decorations;
} MwmHints;

typedef struct _MindowSlot {
   Window mindow;
   CompWindow *window;
   
   void *next;      
} MindowSlot;

		    
typedef enum _MiniwinAction
{
   //basic functionality
   MiniwinActionHideMindow,
   MiniwinActionWindowHideComplete,
   MiniwinActionHideWindow,
   MiniwinActionShowWindow,
   MiniwinActionShowWindowComplete,
   MiniwinActionShowMindow,
   MiniwinActionHoverOn,
   MiniwinActionHoverOff,
   MiniwinActionCheckMindow,
   MiniwinActionRebindMindow,
   
   //mindow creation/destruction
   MiniwinActionCreateMindow,
   MiniwinActionInitializeMindow,
   MiniwinActionDestroyMindow,
   
   //dock events
   MiniwinActionCreateDock,
   MiniwinActionInitializeDock,
   MiniwinActionHideDock,
   MiniwinActionShowDock,
   MiniwinActionUpdateDockBackground,
   
   //taskbar events
   MiniwinActionMinimize,
   MiniwinActionMaximize,
   MiniwinActionRearrangeTaskbar,
   
   //other events
   MiniwinActionHiddenParentGotFocus,
   
} MiniwinAction;

typedef struct _MiniwinEvent
{
   CompWindow *window;
   MiniwinAction action;

   void *next;
   
} MiniwinEvent;


typedef struct _MiniwinDisplay
{
   int screenPrivateIndex;
   HandleEventProc handleEvent;
} MiniwinDisplay;

typedef struct _MiniwinScreen
{
   int windowPrivateIndex;
   
   CompWindow *dockWindow;
   
   //General mindow settings
   unsigned int animationTime;
   unsigned int mindowWidth;
   unsigned int mindowOpacity;
   float scaleOnHover;
   unsigned int opacityOnHover;
   Bool mindowsAlwaysOnTop;
   Bool mindowsOnAllWorkspaces;
   
   //Taskbar settings
   Bool useAsTaskbar;
   int taskbarSpacing;
   Bool showDock;
   unsigned int dockOpacity;
   Bool dockHideOnClick;
   Bool upOnly;
      
   //Internal settings
   Bool runsOnAiglx;   
   unsigned int nSlotsPerRow;
   unsigned int pendingMindows;
   
   Window dock;
   MindowSlot *taskbar;
   MindowSlot *mindows;
   MiniwinEvent *eventQueue;
   int nEvents;   
   Bool hover;
   
   CompOption opt[MINIWIN_SCREEN_OPTION_NUM];
   int wMask;

   PreparePaintScreenProc preparePaintScreen;
   PaintScreenProc paintScreen;
   PaintWindowProc        paintWindow;

} MiniwinScreen;

typedef struct _MiniwinWindow {
    MiniwinWindowState state;   
    
    CompWindow *parent;
    CompWindow *icon;
        
    Bool hover;

    double step;
    
    int originalX;
    int originalY;
    
    int beforeTaskbarX;
    int beforeTaskbarY;
    
    int xMoveDist;
    int yMoveDist;
    
} MiniwinWindow;


typedef struct _ReturnHelper
{
   Window id;
   CompWindow *window;
} ReturnHelper;

//Slot functions
static MindowSlot* addMindowSlot (MindowSlot *slot);
static MindowSlot *getLastSlot (MindowSlot *first);
static int numOccupiedSlots (MindowSlot *first);
static void removeAllSlots (MindowSlot *current);
static MindowSlot *getFreeSlot (MindowSlot *first);
static MindowSlot *getSlotOf (MindowSlot *first, Window id);
static void miniwinRearrangeSlots (MindowSlot *first);

//Taskbar functions
static void miniwinRearrangeTaskbar (CompScreen *screen);
static void miniwinShowDock (CompWindow *dock);
static void miniwinCreateDock (CompScreen *s);
static void miniwinInitializeDock (CompWindow *dock);

//Basic functions
static void miniwinInitializeMindow (CompWindow *mini);
static void miniwinChangeState (CompWindow *window, MiniwinWindowState state);
static void miniwinHide (CompWindow *window);
static void miniwinShow (CompWindow *window);
static void miniwinDestroyMindow (CompWindow *mindow);
static void miniwinCheckMindowState (CompWindow *mini);

static CompWindow *miniwinGetIcon(CompWindow *window);
static CompWindow *miniwinGetWindow (CompWindow *mindow);

static Bool miniwinNoInputFocus (CompWindow *window);
static Bool miniwinBindPixmap (CompWindow *mindow, CompWindow *window);
static Bool miniwinCreateMindow (CompWindow *from_parent);

//Event Handling
static void miniwinPushEvent (CompWindow *window, MiniwinAction action);


//Implementation: taskbar functions / taskbar display functions

static MindowSlot* 
addMindowSlot (MindowSlot *slot)
{
   assert (slot);
   
   MindowSlot *new_slot = malloc (sizeof(MindowSlot));
   new_slot->next = 0;
   new_slot->mindow = 0;
   new_slot->window = 0;
   
   while (slot->next) slot = slot->next;
   slot->next = new_slot;
   
   return new_slot;
}

static MindowSlot *
getLastSlot (MindowSlot *first)
{
   assert (first);
   while (first->next) first = first->next;
   return first;
}

static int
numOccupiedSlots (MindowSlot *first)
{
   assert (first);
   int count = 0;
   do
   {
      if (first->mindow) count++;
   } while (first = first->next);
   return count;
}

static void
removeAllSlots (MindowSlot *current)
{
   assert (current);
   MindowSlot *next;
   do
   {
      next = current->next;
      free (current);            
   } while (current = next);
}

static MindowSlot *
getFreeSlot (MindowSlot *first)
{
   assert (first);
   MindowSlot *current = first;
   do
   {
      if (current->mindow == 0) return current;
   } while (current = current->next);
   return addMindowSlot (getLastSlot(first));
}

static MindowSlot *
getSlotOf (MindowSlot *first, Window id)
{
   do
   {
      if (first->mindow == id) return first;
   } while (first = first->next);
   return 0;
}

static void
miniwinRaiseMindows (MindowSlot *first, CompDisplay *d)
{
   if (first->window)
   {
        MINIWIN_SCREEN(first->window->screen);
        if (ms->dock)
        {
          raiseWindow (findWindowAtDisplay(d, ms->dock));
        }
   }
   do
   {
      if (first->mindow)
         raiseWindow (findWindowAtDisplay(d, first->mindow));
   } while (first = first->next);
}

static void
miniwinRearrangeTaskbar (CompScreen *screen)
{
   MINIWIN_SCREEN (screen);
   
   MindowSlot *current = ms->taskbar;
   miniwinRearrangeSlots (current);
   int index;
            
   int offset = ms->taskbarSpacing / 2;
   int nMinis = numOccupiedSlots (current);
   int hiddenOffset = 0;
               
   int count = 0;
   int row = 1;
   int nRows = nMinis / ms->nSlotsPerRow + 1;
   
   fprintf (stderr, "Dock: Showing %i mindows in %i rows (%i mindows per row)\n", nMinis, nRows, ms->nSlotsPerRow);

   if (ms->showDock && ms->dock)
   {
      CompWindow *dock = findWindowAtScreen(screen, ms->dock); 
      
      int dockWidth;
      if (nRows == 1)
         dockWidth = (ms->mindowWidth * 2 + ms->taskbarSpacing) * nMinis + ms->taskbarSpacing;
      else
         dockWidth = ms->mindowWidth * 2 * ms->nSlotsPerRow;   
   
      XResizeWindow (screen->display->display, dock->id, dockWidth, dock->height);
      XFlush (screen->display->display);
   
      MINIWIN_WINDOW (dock);
      if (mw->state == MINIWIN_WINDOW_STATE_DOCK_HIDDEN && ms->dockHideOnClick)
      {
         hiddenOffset = 1024;
      }
      
      Display *dpy = dock->screen->display->display;
    
      if (nMinis == 0)
         XMoveWindow (dpy, dock->id, (screen->width - dockWidth) / 2, screen->height);   
      else if (mw->state == MINIWIN_WINDOW_STATE_DOCK_HIDDEN && ms->dockHideOnClick)
         XMoveWindow (dpy, dock->id, (screen->width - dockWidth) / 2, screen->height - dock->height/10+2);   
      else
         XMoveWindow (dpy, dock->id, (screen->width - dockWidth) / 2, screen->height - dock->height);
    
      miniwinShowDock (dock);
   }
      
   for (index = 0; current; current = current->next, ++index)
   {
      if (!current->mindow) continue;
      CompWindow *mindow = findWindowAtDisplay (screen->display, current->mindow);
      assert (mindow);
      
      int y = screen->height - row * mindow->height - (row%2) * offset + hiddenOffset;

      int nMindowsInRow = row == nRows ? nMinis % ms->nSlotsPerRow : ms->nSlotsPerRow;
      int rowBegin = ( screen->width - (nMindowsInRow)*(ms->mindowWidth*2 + ms->taskbarSpacing) ) / 2 + offset;
      int x = rowBegin + count * ( ms->mindowWidth * 2 + ms->taskbarSpacing ) 
              + ((row-1)%2) * (ms->mindowWidth - ms->taskbarSpacing);
            
      XMoveWindow (screen->display->display, current->mindow, x, y);
      
      if (++count == ms->nSlotsPerRow)
      {
         count = 0;
         row ++;
      } 
   }    
   
   
}

static void
miniwinRearrangeSlots (MindowSlot *first)
{
   do
   {
      if (!first->mindow && first->next)
      {
         first->mindow = ((MindowSlot*)first->next)->mindow;
         first->window = ((MindowSlot*)first->next)->window;
         ((MindowSlot*)first->next)->mindow = 0;
         ((MindowSlot*)first->next)->window = 0;
      }
   } while (first = first->next);
}


static void
miniwinShowDock (CompWindow *dock)
{        
   CompScreen *screen = dock->screen;
   MINIWIN_SCREEN (screen);
   MINIWIN_WINDOW (dock);
   
   Display *dpy = dock->screen->display->display;
      
   XSelectInput (dpy, dock->id, EnterWindowMask | LeaveWindowMask | KeymapStateMask | ButtonPressMask);
   XFlush (dpy);
   
   dock->state = CompWindowStateHiddenMask;
   dock->state |= CompWindowStateAboveMask;
   dock->state |= CompWindowStateStickyMask;
   dock->paint.opacity = (OPAQUE * ms->dockOpacity) / 100.0f;
   dock->type = CompWindowTypeSplashMask;
   
   miniwinPushEvent (dock, MiniwinActionUpdateDockBackground);       
}


static Visual *
findArgbVisual (Display *dpy, int scr)
{
    XVisualInfo		*xvi;
    XVisualInfo		template;
    int			nvi;
    int			i;
    XRenderPictFormat	*format;
    Visual		*visual;

    template.screen = scr;
    template.depth  = 32;
    template.class  = TrueColor;

    xvi = XGetVisualInfo (dpy,
			  VisualScreenMask |
			  VisualDepthMask  |
			  VisualClassMask,
			  &template,
			  &nvi);
    if (!xvi)
	return 0;

    visual = 0;
    for (i = 0; i < nvi; i++)
    {
	format = XRenderFindVisualFormat (dpy, xvi[i].visual);
	if (format->type == PictTypeDirect && format->direct.alphaMask)
	{
	    visual = xvi[i].visual;
	    break;
	}
    }

    XFree (xvi);

    return visual;
}

static void
miniwinCreateDock (CompScreen *s)
{
   MINIWIN_SCREEN (s);
   
   Atom mwmHintsAtom;
	MwmHints mwmHints;
	 
   Display *dpy = XOpenDisplay(DisplayString(s->display->display));
   
   int whiteColor = WhitePixel(dpy, DefaultScreen(dpy));

   int width = ms->mindowWidth * 2;                     //multiply the desired size with two. don't know why, but other-
   int height = ms->mindowWidth * 1.25f; //wise the icons get too small, which one wouldn't expect.   

   /*Window win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
                                    width, height, 0, whiteColor, whiteColor);

   mwmHintsAtom = XInternAtom (dpy, "_MOTIF_WM_HINTS", 0);
   memset (&mwmHints, 0, sizeof (mwmHints));  
   mwmHints.flags	= MwmHintsDecorations;
   mwmHints.decorations = 0;
   
 	XChangeProperty (dpy, win, mwmHintsAtom, mwmHintsAtom,
			 8, PropModeReplace, (unsigned char *) &mwmHints,
			 sizeof (mwmHints));
   
   Atom type = XInternAtom (dpy, "_NET_WM_WINDOW_TYPE_SPLASH", 0);
   XChangeProperty (dpy, win, XInternAtom (dpy, "_NET_WM_WINDOW_TYPE", 0),
                    XA_ATOM, 32, PropModeReplace, (unsigned char *) &type, 1);
                      
   XMapWindow (s->display->display, win);
   XSelectInput (dpy, win, StructureNotifyMask);
*/
	XSetWindowAttributes attr;
	Atom state[4];
	int nState = 0;
	XWMHints	     xwmh;

	xwmh.flags = InputHint;
	xwmh.input = 0;
			 
   Visual *visual = findArgbVisual (dpy, s->screenNum);
	if (!visual)
	  return;
	  
	attr.background_pixel = 0;
	attr.border_pixel     = 0;
	attr.colormap	      = XCreateColormap (dpy, s->root, visual,
						 AllocNone);

	    
   Window win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
                                    width, height, 0, whiteColor, whiteColor);

	mwmHintsAtom = XInternAtom (dpy, "_MOTIF_WM_HINTS", 0);

	memset (&mwmHints, 0, sizeof (mwmHints));

	mwmHints.flags = MwmHintsDecorations;
	mwmHints.decorations = 0;

	XChangeProperty (dpy, win, mwmHintsAtom, mwmHintsAtom,
			 8, PropModeReplace, (unsigned char *) &mwmHints,
			 sizeof (mwmHints));

	state[nState++] = s->display->winStateAboveAtom;
	state[nState++] = s->display->winStateStickyAtom;
	state[nState++] = s->display->winStateSkipTaskbarAtom;
	state[nState++] = s->display->winStateSkipPagerAtom;

	XChangeProperty (dpy, win,
			 XInternAtom (dpy, "_NET_WM_STATE", 0),
			 XA_ATOM, 32, PropModeReplace,
			 (unsigned char *) state, nState);
        
   Atom type = XInternAtom (dpy, "_NET_WM_WINDOW_TYPE_SPLASH", 0);
   XChangeProperty (dpy, win,
                    XInternAtom (dpy, "_NET_WM_WINDOW_TYPE", 0),
                    XA_ATOM, 32, PropModeReplace,
                    (unsigned char *) &type, 1);
                    
   XMapWindow (s->display->display, win);
   XSelectInput (dpy, win, StructureNotifyMask);
   
   ms->dock = win;
}

static void
miniwinInitializeDock (CompWindow *dock)
{
   assert (dock);
   CompScreen *screen = dock->screen;
   MINIWIN_SCREEN (screen);

   Atom type;   
 	Atom		     mwmHintsAtom;
	MwmHints	     mwmHints;
	 
   Display *dpy = dock->screen->display->display;
   assert(dpy);   
   
   mwmHintsAtom = XInternAtom (dpy, "_MOTIF_WM_HINTS", 0);
	memset (&mwmHints, 0, sizeof (mwmHints));  
   mwmHints.flags	= MwmHintsDecorations;
	mwmHints.decorations = 0;

	XChangeProperty (dpy, dock->id, mwmHintsAtom, mwmHintsAtom,
			 8, PropModeReplace, (unsigned char *) &mwmHints,
			 sizeof (mwmHints));
   type = XInternAtom (dpy, "_NET_WM_WINDOW_TYPE_SPLASH", 0);
   XChangeProperty (dpy, dock->id, XInternAtom (dpy, "_NET_WM_WINDOW_TYPE", 0),
                    XA_ATOM, 32, PropModeReplace, (unsigned char *) &type, 1);
                      
   ms->mindowsAlwaysOnTop = TRUE;
   
   MINIWIN_WINDOW (dock);
   mw->state = MINIWIN_WINDOW_STATE_DOCK_NONE;
   ms->dock = dock->id;
}

/*static Bool
miniwinBindPixmapCropped (CompWindow *mindow, CompWindow *window)
{
   CompScreen *screen = window->screen;
   XVisualInfo  *visinfo;
   unsigned int target;
   
   visinfo = screen->glxPixmapVisuals[window->attrib.depth];
   if (!window->texture.pixmap)
     bindWindow(window);
   mindow->texture.pixmap = glXCreateGLXPixmap (screen->display->display, visinfo, window->pixmap);
   screen->queryDrawable (screen->display->display, window->texture.pixmap, GLX_TEXTURE_TARGET_EXT, &target);
   
   switch (target) {
   case GLX_TEXTURE_2D_EXT:
      mindow->texture.target = GL_TEXTURE_2D;
      mindow->texture.matrix.xx = 1.0f;
      mindow->texture.matrix.yy = -1.0f;
      mindow->texture.matrix.y0 = 1.0f;
      break;
   case GLX_TEXTURE_RECTANGLE_EXT:
      mindow->texture.target = GL_TEXTURE_RECTANGLE_ARB;
      mindow->texture.matrix.xx = 1.0f;
            
      {
         MINIWIN_SCREEN (screen);
         if (ms->runsOnAiglx)
         { 
            mindow->texture.matrix.yy = 1.0f;
            mindow->texture.matrix.y0 = 0.0f;
         }
         else
         {
            mindow->texture.matrix.yy = -1.0f;
            mindow->texture.matrix.y0 = window->height;
         }
      }

      break;
   default:   
      DEBUG("GL: missing GL_TEXTURE_TARGET_EXT -- miniwin won't work :(\n");
      fprintf(stderr,"Return:%d\n",target);
      glXDestroyGLXPixmap (screen->display->display, mindow->texture.pixmap);
      mindow->texture.pixmap = None;
      return FALSE;
   }

   glGenTextures (1, &mindow->texture.name);
   glBindTexture (mindow->texture.target, mindow->texture.name);
   screen->bindTexImage (screen->display->display, mindow->texture.pixmap, GLX_FRONT_LEFT_EXT);
   
   return TRUE;
}*/

static Bool
miniwinBindPixmap (CompWindow *mindow, CompWindow *window)
{
   CompScreen *screen = window->screen;
   //XVisualInfo  *visinfo;
   unsigned int target;
   CompFBConfig *config = &screen->glxPixmapFBConfigs[window->attrib.depth];
   int          attribs[] = {
     GLX_TEXTURE_FORMAT_EXT, config->textureFormat,
     GLX_MIPMAP_TEXTURE_EXT, config->mipmap,
     None
   };

   //visinfo = screen->glxPixmapVisuals[window->attrib.depth];
   if (!window->texture.pixmap)
     bindWindow(window);
   //mindow->texture.pixmap = glXCreateGLXPixmap (screen->display->display, visinfo, window->pixmap);
   mindow->texture.pixmap = (*screen->createPixmap) (screen->display->display,config->fbConfig, window->pixmap, attribs);
   screen->queryDrawable (screen->display->display, window->texture.pixmap, GLX_TEXTURE_TARGET_EXT, &target);
   
   if ((!mindow->mapNum && mindow->attrib.map_state == IsViewable)) return;
      
   switch (target) {
   case GLX_TEXTURE_2D_EXT:
      mindow->texture.target = GL_TEXTURE_2D;
      DEBUG("GL: GL_TEXTURE_2D\n");
      mindow->texture.matrix.xx = 1.0f / mindow->width;
      if (config->yInverted)
      {
        mindow->texture.matrix.yy = 1.0f / mindow->height;
        mindow->texture.matrix.y0 = 0.0f;
      }
      else
      {
        mindow->texture.matrix.yy = -1.0f / mindow->height;
        mindow->texture.matrix.y0 = 1.0f;
      }
      break;
   case GLX_TEXTURE_RECTANGLE_EXT:
      mindow->texture.target = GL_TEXTURE_RECTANGLE_ARB;
      DEBUG("GL: GL_TEXTURE_RECTANGLE_ARB\n");
      mindow->texture.matrix.xx = (float)window->width/(float)mindow->width;
            
      {
         if (config->yInverted)
         { 
            mindow->texture.matrix.yy = ((float)window->height/(float)mindow->height);
            mindow->texture.matrix.y0 = 0.0f;
         }
         else
         {
            mindow->texture.matrix.yy = -((float)window->height/(float)mindow->height);
            mindow->texture.matrix.y0 = window->height;
         }
      }

      break;
   default:   
      DEBUG("GL: missing GL_TEXTURE_TARGET_EXT -- miniwin won't work :(\n");
      fprintf(stderr,"Return:%d\n",target);
      glXDestroyGLXPixmap (screen->display->display, mindow->texture.pixmap);
      mindow->texture.pixmap = None;
      return FALSE;
   }

   glGenTextures (1, &mindow->texture.name);
   glBindTexture (mindow->texture.target, mindow->texture.name);
   screen->bindTexImage (screen->display->display, mindow->texture.pixmap, GLX_FRONT_LEFT_EXT, NULL);
   
   return TRUE;
}


static Bool 
miniwinCreateMindow (CompWindow *from_parent)
{  
   CompWindow *orig = from_parent;
   CompScreen *screen = orig->screen;

   MINIWIN_SCREEN (screen);
   MINIWIN_WINDOW (from_parent);
   
   Atom type;   
 	Atom		     mwmHintsAtom;
	MwmHints	     mwmHints;
   
   Display *dpy = XOpenDisplay(DisplayString(screen->display->display));
   assert(dpy);
  
   int blackColor = BlackPixel(dpy, DefaultScreen(dpy)); //just playing around
   int whiteColor = WhitePixel(dpy, DefaultScreen(dpy));

   int width = ms->mindowWidth * 2;                     //multiply the desired size with two. don't know why, but other-
   int height = (width * orig->height) / orig->width; //wise the icons get too small, which one wouldn't expect.
   
   //don't let the mindow get too long.
   if (height > ms->mindowWidth * 2)
   {
      height = ms->mindowWidth * 2;
      width = (height * orig->width) / orig->height;
   }
  
   DEBUG ("Trying to create a simple mindow\n"); 
   Window win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
                                    width, height, 0, whiteColor, blackColor);

   
   mwmHintsAtom = XInternAtom (dpy, "_MOTIF_WM_HINTS", 0);
	memset (&mwmHints, 0, sizeof (mwmHints));  
   mwmHints.flags	= MwmHintsDecorations;
	mwmHints.decorations = 0;

	XChangeProperty (dpy, win, mwmHintsAtom, mwmHintsAtom,
			 8, PropModeReplace, (unsigned char *) &mwmHints,
			 sizeof (mwmHints));

   type = XInternAtom (dpy, "_NET_WM_WINDOW_TYPE_SPLASH", 0);
   XChangeProperty (dpy, win, XInternAtom (dpy, "_NET_WM_WINDOW_TYPE", 0),
                    XA_ATOM, 32, PropModeReplace, (unsigned char *) &type, 1);
                      
   XMapWindow (screen->display->display, win);
   XSelectInput (dpy, win, StructureNotifyMask);
 
   mw->state = MINIWIN_WINDOW_STATE_NORMAL_WAITING;
   mw->originalX = orig->serverX;
   mw->originalY = orig->serverY;
 
   miniwinRearrangeSlots (ms->mindows);
   MindowSlot *new_mindow = getFreeSlot(ms->mindows);
   new_mindow->mindow = win;
   new_mindow->window = from_parent;
   
   return TRUE;
}

static void
miniwinInitializeMindow (CompWindow *mini)
{
   CompScreen *screen = mini->screen;
   
   Atom mwmHintsAtom = XInternAtom (screen->display->display, "_MOTIF_WM_HINTS", 0);
         
   MwmHints mwmHints;
   memset (&mwmHints, 0, sizeof (mwmHints));  
   mwmHints.flags	= MwmHintsDecorations;
   mwmHints.decorations = 0;

   XChangeProperty (screen->display->display, mini->id, mwmHintsAtom, mwmHintsAtom,
                    8, PropModeReplace, (unsigned char *) &mwmHints, sizeof (mwmHints));
      	                 
   Atom type = XInternAtom (screen->display->display, "_NET_WM_WINDOW_TYPE_SPLASH", 0);
   XChangeProperty (screen->display->display, mini->id, 
                    XInternAtom (screen->display->display, "_NET_WM_WINDOW_TYPE", 0),
                    XA_ATOM, 32, PropModeReplace, (unsigned char *) &type, 1);
                          
   XSelectInput (screen->display->display, mini->id, 
                 EnterWindowMask | LeaveWindowMask | FocusChangeMask | StructureNotifyMask
                 | VisibilityChangeMask);
   
   MINIWIN_WINDOW (mini);   
   mw->originalX = -1;
   mw->originalY = -1;
}

static void
miniwinChangeState (CompWindow *window, MiniwinWindowState state)
{
   MINIWIN_WINDOW (window);
   mw->state = state;
}

static void
miniwinHide (CompWindow *window)
{
   MINIWIN_WINDOW (window);
   mw->originalX = window->serverX;
   mw->originalY = window->serverY;
   XMoveWindow (window->screen->display->display, window->id, window->screen->width/2, window->screen->height*4);   
   mw->hover = FALSE;
   MINIWIN_SCREEN (window->screen);
   ms->hover = FALSE;
}

static void
miniwinShow (CompWindow *window)
{
   MINIWIN_WINDOW (window);
   XMoveWindow (window->screen->display->display, window->id, mw->originalX, mw->originalY);   
   raiseWindow (window);
   updateWindowAttributes (window,TRUE);   
   moveInputFocusToWindow (window);
}

static CompWindow *
miniwinGetIcon(CompWindow *window)
{
   MINIWIN_WINDOW (window);
   assert (mw->icon);
   return mw->icon;
}

static CompWindow *
miniwinGetWindow (CompWindow *mindow)
{
   MINIWIN_WINDOW (mindow);
   assert (mw->parent);
   return mw->parent;
}

static void
miniwinDestroyMindow (CompWindow *mindow)
{
   MINIWIN_SCREEN (mindow->screen);
   if (ms->useAsTaskbar)
   {
      MindowSlot *slot = getSlotOf (ms->taskbar, mindow->id);
      if (slot) slot->mindow = 0;
      miniwinRearrangeTaskbar (mindow->screen);
   }  
   
   XDestroyWindow (mindow->screen->display->display, mindow->id);
   destroyWindow (mindow);
}

static void
miniwinCheckMindowState (CompWindow *mini)
{
   MINIWIN_SCREEN (mini->screen);
   mini->state = CompWindowStateHiddenMask; //TODO: find another way - this hacks around the trailfocus plugin.
   if (ms->mindowsAlwaysOnTop)
      mini->state |= CompWindowStateAboveMask;
   if (ms->mindowsOnAllWorkspaces)
      mini->state |= CompWindowStateStickyMask;
      
   mini->paint.saturation = OPAQUE;
   mini->paint.brightness = OPAQUE;
   MINIWIN_WINDOW (mini);
   if (mw->hover)
      mini->paint.opacity = (OPAQUE * ms->opacityOnHover) / 100.0f;
   else
      mini->paint.opacity = (OPAQUE * ms->mindowOpacity) / 100.0f;
}

static Bool
miniwinNoInputFocus (CompWindow *window)
{
   MINIWIN_WINDOW (window);
   return mw->state == MINIWIN_WINDOW_STATE_MINDOW_ICON || mw->state == MINIWIN_WINDOW_STATE_MINDOW_TASKBAR;
}

//Event handling

static void
miniwinPushEvent (CompWindow *window, MiniwinAction action)
{
   MINIWIN_SCREEN (window->screen);
   DEBUG ("Event added!\n");
   MiniwinEvent *queue = ms->eventQueue;
   if (!queue)
   {
      queue = ms->eventQueue = malloc (sizeof(MiniwinEvent));
      queue->next = 0;
      queue->window = window;
      queue->action = action;
   }
   else
   {
      while (queue->next)
      {
         queue = queue->next;
      }
      queue->next = malloc (sizeof(MiniwinEvent));
      queue = queue->next;
      queue->next = 0;
      queue->window = window;
      queue->action = action;
   }
}

static Bool 
miniwinPendingEvent (CompScreen *screen)
{
   MINIWIN_SCREEN (screen);
   if (ms->eventQueue) return TRUE;
   return FALSE;
}

static MiniwinEvent
miniwinPopEvent (CompScreen *screen)
{
   assert (miniwinPendingEvent (screen));
   MINIWIN_SCREEN (screen);
   
   MiniwinEvent *queue = ms->eventQueue;
   MiniwinEvent *prev = 0;
   while (queue->next)
   {
      prev = queue;
      queue = queue->next;
   }
   MiniwinEvent retval = *queue;   
   free (queue);
   
   if (prev) prev->next = 0;
   else ms->eventQueue = 0;   
   
   return retval;
}

//Main program

static void
miniwinIconify (CompWindow* window)
{ 
   CompWindow *mindow = miniwinGetIcon (window);
   MINIWIN_WINDOW (mindow);
   if (mw->originalX == -1 && mw->originalY == -1)
   {
      mw->originalX = window->serverX;
      mw->originalY = window->serverY;
   }
   
   DEBUG ("Sending Mindowing event!\n");
   miniwinPushEvent (mindow, MiniwinActionShowMindow);
   miniwinPushEvent (window, MiniwinActionHideWindow);
}

static void
miniwinDeiconify (CompWindow* mindow)
{        
   CompWindow *window = miniwinGetWindow (mindow);
      
   DEBUG ("Sending Demindowing event!\n");
   miniwinPushEvent (window, MiniwinActionShowWindow);
   miniwinPushEvent (mindow, MiniwinActionHideMindow);
}

static void
miniwinUpdate (MiniwinEvent event, const WindowPaintAttrib *attrib, unsigned int mask)
{
   switch (event.action)
   {
   case MiniwinActionHideWindow:
      DEBUG ("HideWindow event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_NORMAL_HIDING);
      {
         MINIWIN_WINDOW (event.window);
         mw->step = 100;
      }
      break;
   case MiniwinActionWindowHideComplete:
      DEBUG ("WindowHideComplete event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_NORMAL_HIDDEN);
      (*event.window->screen->setWindowScale) (event.window, 1.0f, 1.0f);
      //HACK:
      {
         MINIWIN_WINDOW (event.window);
         event.window->serverX = mw->originalX;
         event.window->serverY = mw->originalY;
      }
      //: END HACK
      miniwinHide (event.window);
      break;
   case MiniwinActionHideMindow:
      DEBUG ("HideMindow event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_MINDOW_HIDDEN);
      miniwinHide (event.window);
      break;
   case MiniwinActionShowWindow:
      DEBUG ("ShowWindow event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_NORMAL_UNHIDING);
      {
         MINIWIN_WINDOW (event.window);
         mw->step = 100;
      }
      break;
   case MiniwinActionShowWindowComplete:
      DEBUG ("ShowWindowComplete event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_NORMAL_NONE);
      miniwinShow (event.window);
      (*event.window->screen->setWindowScale) (event.window, 1.0f, 1.0f);
      break;
   case MiniwinActionShowMindow:
      DEBUG ("ShowMindow event received!\n");
      miniwinPushEvent (event.window, MiniwinActionCheckMindow);
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_MINDOW_ICON);
      miniwinShow (event.window);
      break;

   case MiniwinActionCreateMindow:
      DEBUG ("CreateMindow event received!\n");
      miniwinCreateMindow (event.window);
      break;
   case MiniwinActionInitializeMindow:
      DEBUG ("InitializeMindow event received!\n");      
      miniwinInitializeMindow (event.window);
      miniwinHide (event.window);   
      {
         MINIWIN_WINDOW (event.window);
         mw->originalX = -1;
         mw->originalY = -1;
      }    
      miniwinPushEvent (event.window, MiniwinActionCheckMindow);
      break;
   case MiniwinActionCheckMindow:
      DEBUG ("CheckMindow event received!\n");
      {
         MINIWIN_SCREEN (event.window->screen);
         MINIWIN_WINDOW (event.window);
         CompWindow *parent = mw->parent;
         
         int width = ms->mindowWidth * 2;                  
         int height = (width * parent->height) / parent->width;
   
         if (height > ms->mindowWidth * 2)
         {
            height = ms->mindowWidth * 2;
            width = (height * parent->width) / parent->height;
         }
         XResizeWindow (event.window->screen->display->display, event.window->id, width, height);   
         miniwinCheckMindowState (event.window);   
         miniwinPushEvent (event.window, MiniwinActionRebindMindow);  
      }
      break;
   case MiniwinActionRebindMindow:
      {
         MINIWIN_WINDOW (event.window);
         miniwinBindPixmap (event.window, mw->parent);
         (*event.window->screen->drawWindowTexture) (event.window, &event.window->texture, &attrib, mask);
      }
      break;
   case MiniwinActionDestroyMindow:
      DEBUG ("DestroyMindow event received!\n");
      miniwinDestroyMindow (event.window);
      break;
   case MiniwinActionHoverOn:
      DEBUG ("HoverOn event received!\n");
      {
         MINIWIN_SCREEN (event.window->screen);
         MINIWIN_WINDOW (event.window);
         raiseWindow (event.window);
         if (!mw->hover)
         {
                if (ms->scaleOnHover != 1.0f)
                {
                (*event.window->screen->setWindowScale)(event.window, ms->scaleOnHover, ms->scaleOnHover);
        
                mw->xMoveDist=-(event.window->width*ms->scaleOnHover-event.window->width)/2;
                mw->yMoveDist=-(event.window->height*ms->scaleOnHover-event.window->height);
                
                if (!ms->upOnly) mw->yMoveDist/=2;
        
                moveWindow(event.window,mw->xMoveDist,mw->yMoveDist, TRUE, TRUE);
                }                             
                
                mw->hover = TRUE;
         }
         ms->hover = TRUE;
         event.window->paint.opacity = (OPAQUE * ms->opacityOnHover) / 100;
         addWindowDamage(event.window);
      }
      break;
   case MiniwinActionHoverOff:
      DEBUG ("HoverOff event received!\n");
      {
         MINIWIN_WINDOW (event.window);
         if (mw->hover)
         {
            MINIWIN_SCREEN (event.window->screen);
            (*event.window->screen->setWindowScale)(event.window, 1.0f, 1.0f);
            moveWindow(event.window,-mw->xMoveDist,-mw->yMoveDist, TRUE, TRUE);
            event.window->paint.opacity = (OPAQUE * ms->mindowOpacity) / 100;
            mw->hover = FALSE;
            ms->hover = FALSE;
            addWindowDamage(event.window);
         }
      }
      break;
   case MiniwinActionCreateDock:
      DEBUG ("CreateDock event received!\n");
      {
         miniwinCreateDock (event.window->screen); //attention: the event.window here is _NOT_ the dock.
      }
      break;
   case MiniwinActionInitializeDock:
      DEBUG ("InitializeDock event received!\n");
      {
         miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_DOCK_NONE);
         miniwinRearrangeTaskbar (event.window->screen);
         miniwinInitializeDock (event.window);
      }
      break;
   case MiniwinActionHideDock:
      DEBUG ("HideDock event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_DOCK_HIDDEN);
      miniwinRearrangeTaskbar (event.window->screen);
      break;
   case MiniwinActionShowDock:
      DEBUG ("ShowDock event received!\n");
      miniwinChangeState (event.window, MINIWIN_WINDOW_STATE_DOCK_NONE);
      miniwinRearrangeTaskbar (event.window->screen);
      {
         MINIWIN_SCREEN (event.window->screen);            
         miniwinRaiseMindows (ms->taskbar, event.window->screen->display);
      }
      break;
   case MiniwinActionUpdateDockBackground:
      {
         MINIWIN_SCREEN (event.window->screen);            
         if (ms->dockWindow)
         {
            DEBUG ("Binding a window to the dock!\n");
            miniwinBindPixmap (event.window, ms->dockWindow);
         }
      }
      break;
   case MiniwinActionHiddenParentGotFocus:
      DEBUG ("HiddenParentGotFocus event received!\n");
      {
        miniwinPushEvent (event.window, MiniwinActionMaximize);
      }
      break;
   case MiniwinActionMinimize:
      DEBUG ("Minimize event received!\n");
      {
         MINIWIN_SCREEN (event.window->screen);
         MINIWIN_WINDOW (event.window);
                  
         MindowSlot *slot = getFreeSlot (ms->taskbar);
         slot->mindow = mw->icon->id;
         
         CompWindow *icon = mw->icon;
         {
            MINIWIN_WINDOW (icon);         
            mw->beforeTaskbarX = mw->originalX;
            mw->beforeTaskbarY = mw->originalY;
         }
         
         miniwinPushEvent (mw->icon, MiniwinActionCheckMindow);
         miniwinPushEvent (event.window, MiniwinActionHideWindow);
         miniwinPushEvent (event.window, MiniwinActionRearrangeTaskbar);
         miniwinChangeState (mw->icon, MINIWIN_WINDOW_STATE_MINDOW_TASKBAR);  
      }
      break;
   case MiniwinActionMaximize:
      DEBUG ("Maximize event received!\n");
      {
         MINIWIN_SCREEN (event.window->screen);
         MINIWIN_WINDOW (event.window);        
             
         MindowSlot *slot = getSlotOf (ms->taskbar, event.window->id);
         if (slot)
         {
            slot->mindow = 0;

            mw->originalX = mw->beforeTaskbarX;
            mw->originalY = mw->beforeTaskbarY;

            miniwinPushEvent (event.window, MiniwinActionHideMindow);
            miniwinPushEvent (mw->parent, MiniwinActionShowWindow);
            miniwinPushEvent (event.window, MiniwinActionRearrangeTaskbar);
         }
      }
      break;
   case MiniwinActionRearrangeTaskbar:
      DEBUG ("RearrangeTaskbar event received!\n");
      {
         MINIWIN_SCREEN (event.window->screen);
         miniwinRearrangeTaskbar (event.window->screen);
         miniwinRaiseMindows (ms->taskbar, event.window->screen->display);
      }
      break;
   default:
      DEBUG ("Unhandled event!\n");
   }
}

static void 
miniwinPreparePaintScreen (CompScreen *screen, int msSinceLastPaint)
{
   MINIWIN_SCREEN (screen);
  
   CompWindow *w;
   for (w = screen->windows; w; w = w->next)
	{
	  MINIWIN_WINDOW (w);
	  if (mw->step)
	  {	
        mw->step -= msSinceLastPaint / (10 * ms->animationTime / 1000.0f);
        if (mw->step <= 0)
	     {
	       mw->step = 0;          
	     }
	       
	     if (mw->state == MINIWIN_WINDOW_STATE_NORMAL_HIDING)
	     {     
          if (mw->step == 0)
          {
             miniwinPushEvent (w, MiniwinActionWindowHideComplete);
             return;
          }
	             
          double step = mw->step / 100.0f;

          int diffX = abs (w->width - mw->icon->width);
          int width = mw->icon->width + step * diffX;
          double scaleX = width / (w->width * 1.0f);
        
          int diffY = abs (w->height - mw->icon->height);
          int height = mw->icon->height + step * diffY;
          double scaleY = height / (w->height * 1.0f);
        
          (*screen->setWindowScale) (w, scaleX, scaleY);
        
          int distX = (1.0f - step) * (mw->icon->serverX - w->serverX);
          int distY = (1.0f - step) * (mw->icon->serverY - w->serverY);
        
          int dX = w->serverX + distX;
          int dY = w->serverY + distY;
        
          XMoveWindow (screen->display->display, w->id, dX, dY);
	    }
	    else if (mw->state == MINIWIN_WINDOW_STATE_NORMAL_UNHIDING)
	    {
	       if (mw->step == 0)
          {
             miniwinPushEvent (w, MiniwinActionShowWindowComplete);
             return;
          }          
          double step = 1.0f - (mw->step / 100.0f);

          int diffX = abs (w->width - mw->icon->width);
          int width = mw->icon->width + step * diffX;
          double scaleX = width / (w->width * 1.0f);
        
          int diffY = abs (w->height - mw->icon->height);
          int height = mw->icon->height + step * diffY;
          double scaleY = height / (w->height * 1.0f);
        
          (*screen->setWindowScale) (w, scaleX, scaleY);
          
          MiniwinWindow *tmp = mw;
          {
            MINIWIN_WINDOW (tmp->icon);
            tmp = mw;
          }
          int distX = step * (mw->originalX - tmp->originalX);
          int distY = step * (mw->originalY - tmp->originalY);
        
          int dX = tmp->originalX + distX;
          int dY = tmp->originalY + distY;
        
          XMoveWindow (screen->display->display, w->id, dX, dY);
	    }
            addWindowDamage (w);
            damageScreen (screen);
          }
   }
  
   UNWRAP (ms, screen, preparePaintScreen);
   (*screen->preparePaintScreen) (screen, msSinceLastPaint);
   WRAP (ms, screen, preparePaintScreen, miniwinPreparePaintScreen);
}
static Bool
miniwinPaintScreen (CompScreen              *s,
                      const ScreenPaintAttrib *sAttrib,
                      Region                  region,
                      unsigned int            mask)
{
  Bool status;

  MINIWIN_SCREEN (s);

  if (ms->hover)
    mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
    
  CompWindow *w;
  for (w = s->windows; w; w = w->next)
  {	
     MINIWIN_WINDOW (w);
	  if (mw->step)
	  {
   	  mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
   	  break;
	  }
  }
	  

  UNWRAP (ms, s, paintScreen);
  status = (*s->paintScreen) (s, sAttrib, region, mask);
  WRAP (ms, s, paintScreen, miniwinPaintScreen);

  return status;
}


//now, this is getting interesting, because we are doing a lot of initialization stuff in here.
static Bool
miniwinPaintWindow (CompWindow *w, const WindowPaintAttrib *attrib, Region region, unsigned int mask)
{
   MINIWIN_WINDOW (w);
   Bool status;
   CompScreen *screen = w->screen;
   MINIWIN_SCREEN (screen);  
        
   switch (mw->state)
   {
   case MINIWIN_WINDOW_STATE_DOCK_CREATED:
      miniwinPushEvent (w, MiniwinActionInitializeDock);
      break;
   case MINIWIN_WINDOW_STATE_DOCK_NONE:
      //miniwinRaiseMindows (ms->taskbar, w->screen->display);
      //this took too much cpu.
      break;
   case MINIWIN_WINDOW_STATE_NORMAL_HIDDEN:
     if (mw->icon) addWindowDamage(mw->icon);
     return TRUE;
   case MINIWIN_WINDOW_STATE_MINDOW_HIDDEN: return TRUE;
   case MINIWIN_WINDOW_STATE_NORMAL_HIDING:
   case MINIWIN_WINDOW_STATE_NORMAL_UNHIDING:
      mask |= PAINT_WINDOW_TRANSFORMED_MASK;
      break;
   case MINIWIN_WINDOW_STATE_MINDOW_TASKBAR:
   case MINIWIN_WINDOW_STATE_MINDOW_ICON:
      /*miniwinCheckMindowState (w);
      if (mw->parent->damaged)
      {
         damageScreen (screen);
         addWindowDamage (w);    
   }*/
   default:
      break;
   }
   
   if (mw->hover)
   {
      mask |= PAINT_WINDOW_TRANSFORMED_MASK;
      ms->hover = TRUE;
   }
   
   if (ms->dock == 0 && ms->showDock)
   {
      miniwinPushEvent (w, MiniwinActionCreateDock);
   }
   
   UNWRAP (ms, screen, paintWindow);
   status = (*screen->paintWindow) (w, attrib, region, mask);
   WRAP (ms, screen, paintWindow, miniwinPaintWindow);
   
   while (miniwinPendingEvent (screen))
   {
      DEBUG ("Processing the next event...\n");
      miniwinUpdate (miniwinPopEvent(screen), attrib, mask);
   }
   
   //for each newly created window there will be a mindow: it's "icon"
   //which is available until the window gets destroyed.
   if (mw->state == MINIWIN_WINDOW_STATE_NORMAL_INITIALIZING)
   {
      MindowSlot *minSlot = getSlotOf (ms->mindows, w->id);
      if (minSlot)
      {
         DEBUG ("Trying to initialize a mindow\n");
         mw->state = MINIWIN_WINDOW_STATE_MINDOW_CREATED;
         mw->parent = minSlot->window;
         
         miniwinPushEvent (w, MiniwinActionInitializeMindow);
         
         CompWindow *parent = mw->parent;
         {
            MINIWIN_WINDOW (parent);
            mw->icon = w;     
            mw->state = MINIWIN_WINDOW_STATE_NORMAL_NONE;
         }
      }
      else
      {
         if (ms->wMask & w->type)
         {
            mw->state = MINIWIN_WINDOW_STATE_NORMAL_NONE;
            DEBUG ("Trying to create a mindow from a window\n");
            miniwinPushEvent (w, MiniwinActionCreateMindow);
         }
      }
   }
   
/*   if ((mw->state == MINIWIN_WINDOW_STATE_DOCK_NONE || mw->state == MINIWIN_WINDOW_STATE_DOCK_HIDDEN) 
                && ms->dockWindow)
            {
               miniwinBindPixmapCropped (w, ms->dockWindow);
            } */
   
   return status;
}

//TODO: add some code in order to be able to move iconified windows more easily.
static void
miniwinHandleEvent (CompDisplay* d, XEvent* event)
{
   CompWindow* w;
   MINIWIN_DISPLAY (d);
   
   switch (event->type) {
   case VisibilityNotify:
      w = findWindowAtDisplay (d, event->xvisibility.window);
      if (w)
      {
         MINIWIN_WINDOW (w);
         MINIWIN_SCREEN (w->screen);
         if (mw->parent)
         {
            miniwinPushEvent (w, MiniwinActionRebindMindow);
            miniwinPushEvent (w, MiniwinActionRearrangeTaskbar);
         }
      }
      break;
   case FocusOut:
      w = findWindowAtDisplay (d, event->xfocus.window);
      if (w)
      {
         MINIWIN_SCREEN (w->screen);
         Window focus;
         int tmp;
         XGetInputFocus (d->display, &focus, &tmp);
         if (focus == ms->dock)
         {
            XSetInputFocus (d->display, w->id, RevertToParent, CurrentTime);  
         }
      }
      break;
   case FocusIn:
      w = findWindowAtDisplay (d, event->xfocus.window);
      if (w)
      {
         MINIWIN_WINDOW (w);
         if (mw->state == MINIWIN_WINDOW_STATE_NORMAL_HIDDEN)
         {
            miniwinPushEvent (mw->icon, MiniwinActionHiddenParentGotFocus);
         }
      }
      break;
   //Iconify on minimize:
   case UnmapNotify:
      w = findWindowAtDisplay (d, event->xunmap.window);
      if (w)
      {
         CompScreen *s = w->screen;

         if (!w->minimized) break;
         MINIWIN_SCREEN (s);
         if (!ms->useAsTaskbar) break;      
         
         MINIWIN_WINDOW (w);
         if (mw->state == MINIWIN_WINDOW_STATE_NORMAL_NONE)
         {
            XMapWindow (w->screen->display->display, w->id);
            miniwinPushEvent (w, MiniwinActionMinimize);
         }
      }
      break;
   case EnterNotify:
      w = findWindowAtDisplay (d, event->xcrossing.window);
      if (w)
      {
         MINIWIN_WINDOW (w);
         if (mw->state == MINIWIN_WINDOW_STATE_MINDOW_ICON ||
             mw->state == MINIWIN_WINDOW_STATE_MINDOW_TASKBAR)
         {
            miniwinPushEvent (w, MiniwinActionHoverOn);
         }
         else if (mw->state == MINIWIN_WINDOW_STATE_DOCK_HIDDEN)
         {
            miniwinPushEvent (w, MiniwinActionShowDock);
         }
      }
      break;
   case LeaveNotify:
      w = findWindowAtDisplay (d, event->xcrossing.window);
      if (w)
      {
         MINIWIN_WINDOW (w);
         if (mw->state == MINIWIN_WINDOW_STATE_MINDOW_ICON ||
             mw->state == MINIWIN_WINDOW_STATE_MINDOW_TASKBAR)
         {
            miniwinPushEvent (w, MiniwinActionHoverOff);
         }
      }
      break;
   case DestroyNotify:
      w = findWindowAtDisplay (d, event->xdestroywindow.window);
      if (w)
      {
         MINIWIN_WINDOW (w); 
         if (mw->icon)
         {
            miniwinPushEvent (mw->icon, MiniwinActionDestroyMindow);
         }
         MINIWIN_SCREEN (w->screen);
         if (w->id == ms->dockWindow)
         {
            ms->dockWindow = 0;
         }
      }   
      break;
   case ButtonPress:
   case ButtonRelease:       
      w = findWindowAtDisplay (d, event->xkey.window);
      if (w)
      {         
         CompScreen *s = w->screen;      
         MINIWIN_SCREEN (s);           
         MINIWIN_WINDOW (w);
                  
         if (mw->state == MINIWIN_WINDOW_STATE_DOCK_NONE)
         {
            if (ms->dockHideOnClick)
               miniwinPushEvent (w, MiniwinActionHideDock);
            else
               miniwinRearrangeTaskbar (s);
         }
            
         if (!(ms->wMask & w->type)) break;
      }
   case KeyPress:
   case KeyRelease:
      w = findWindowAtDisplay (d, event->xkey.window);
      if (w)
      {  
         CompScreen *s = w->screen;
         MINIWIN_SCREEN (s);       
         MINIWIN_WINDOW (w);
                    
         if (!(ms->wMask & w->type)) break;
         
         if (eventMatches (d,event,
                     &ms->opt[MINIWIN_SCREEN_OPTION_DOCK_SET_BACKGROUND_KEY]) &&
             mw->state == MINIWIN_WINDOW_STATE_NORMAL_NONE)
         {
            ms->dockWindow = w;
            miniwinPushEvent (w, MiniwinActionRearrangeTaskbar);            
         }
         
         if (eventMatches (d,event,&ms->opt[MINIWIN_SCREEN_OPTION_MINDOW_KEY]) && 
             mw->state == MINIWIN_WINDOW_STATE_NORMAL_NONE)
         {   
            miniwinIconify (w);
         }
         if (eventMatches (d,event,&ms->opt[MINIWIN_SCREEN_OPTION_DEMINDOW_KEY]))
         {
            if (mw->state == MINIWIN_WINDOW_STATE_MINDOW_ICON)
               miniwinDeiconify (w);
            else if (mw->state == MINIWIN_WINDOW_STATE_MINDOW_TASKBAR);
               miniwinPushEvent (w, MiniwinActionMaximize);
         }
      }
      break;
   default:
      break;
   }
   
   UNWRAP (md, d, handleEvent);
   (*d->handleEvent) (d, event);
   WRAP (md, d, handleEvent, miniwinHandleEvent);
}

//From here on, it will get boring. I don't mind if you just close the file now.
//What follows is only compiz initialization stuff...
 
static CompOption*
miniwinGetScreenOptions (CompScreen* screen, int* count)
{
   MINIWIN_SCREEN (screen); //hate this way of getting a miniwin object. have to change that.
   *count = NUM_OPTIONS (ms);
   return ms->opt;
}

static Bool
miniwinSetScreenOption (CompScreen* screen, char* name, CompOptionValue* value)
{
   CompOption* o;
   int index;
   
   MINIWIN_SCREEN (screen);
   
   o = compFindOption (ms->opt, NUM_OPTIONS (ms), name, &index);
   if (!o) return FALSE;
   
   switch (index) {
   case MINIWIN_SCREEN_OPTION_SCALE_ON_HOVER:
      if (compSetFloatOption (o, value))
      {
         ms->scaleOnHover = o->value.f;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_DOCK_OPACITY:
      if (compSetIntOption (o, value))
      {
         ms->dockOpacity = o->value.i;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_TASKBAR_SPACING:
      if (compSetIntOption (o, value))
      {
         ms->taskbarSpacing = o->value.i;
         ms->nSlotsPerRow = screen->width / ( ms->mindowWidth * 2 + ms->taskbarSpacing );
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_ANIMATION_TIME:
      if (compSetIntOption (o, value))
      {
         ms->animationTime = o->value.i;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_OPACITY_ON_HOVER:
      if (compSetIntOption (o, value))
      {
         ms->opacityOnHover = o->value.i;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_MINDOW_OPACITY:
      if (compSetIntOption (o, value))
      {
         ms->mindowOpacity = o->value.i;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_MINDOW_WIDTH:
      if (compSetIntOption (o, value))
      {
         ms->mindowWidth = o->value.i;
         ms->nSlotsPerRow = screen->width / ( ms->mindowWidth * 2 + ms->taskbarSpacing );
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_ONLY_MOVE_UP:
      if (compSetBoolOption (o, value))
      {
         ms->upOnly = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_MINDOWS_ALWAYS_ON_TOP:
      if (compSetBoolOption (o, value))
      {
         ms->mindowsAlwaysOnTop = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_MINDOWS_ON_ALL_WORKSPACES:
      if (compSetBoolOption (o, value))
      {
         ms->mindowsOnAllWorkspaces = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_USE_AS_TASKBAR:
      if (compSetBoolOption (o, value))
      {
         ms->useAsTaskbar = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_SHOW_DOCK:
      if (compSetBoolOption (o, value))
      {
         ms->showDock = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_DOCK_HIDE_ON_CLICK:
      if (compSetBoolOption (o, value))
      {
         ms->dockHideOnClick = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_RUNS_ON_AIGLX:
      if (compSetBoolOption (o, value))
      {
         ms->runsOnAiglx = o->value.b;
         return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_MINDOW_KEY:
   case MINIWIN_SCREEN_OPTION_DEMINDOW_KEY:
   case MINIWIN_SCREEN_OPTION_DOCK_SET_BACKGROUND_KEY:
      {
         removeScreenBinding (screen, &o->value.bind);
         if (compSetBindingOption (o, value)) return TRUE;
      }
      break;
   case MINIWIN_SCREEN_OPTION_WINDOW_TYPE:
      if (compSetOptionList (o, value))
      {
         ms->wMask = compWindowTypeMaskFromStringList (&o->value);
         ms->wMask &= ~CompWindowTypeDesktopMask;
         return TRUE;
      }
   default: 
      break;
   }
   
   return FALSE;
}

static void
miniwinScreenInitOptions (MiniwinScreen* ms)
{
   CompOption* o;
   int i;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_SCALE_ON_HOVER];
   o->name = "scale_on_hover_factor";
   o->shortDesc = "Scale mindow on hover";
   o->longDesc = "Scale mindow on hover (factor)";
   o->type = CompOptionTypeFloat;
   o->value.f = MINIWIN_SCALE_ON_HOVER_DEFAULT;
   o->rest.f.min = MINIWIN_SCALE_ON_HOVER_MIN;
   o->rest.f.max = MINIWIN_SCALE_ON_HOVER_MAX;
   o->rest.f.precision = MINIWIN_SCALE_ON_HOVER_STEP;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_MINDOW_WIDTH];
   o->name = "mindow_width";
   o->shortDesc = "Standard width of mindows";
   o->longDesc = "Standard width of mindows";
   o->type = CompOptionTypeInt;
   o->value.i = MINIWIN_MINDOW_WIDTH_DEFAULT;
   o->rest.i.min = MINIWIN_MINDOW_WIDTH_MIN;
   o->rest.i.max = MINIWIN_MINDOW_WIDTH_MAX;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_ANIMATION_TIME];
   o->name = "animation_time";
   o->shortDesc = "Animation duration";
   o->longDesc = "Duration of mindow animations (in miliseconds)";
   o->type = CompOptionTypeInt;
   o->value.i = MINIWIN_ANIMATION_TIME_DEFAULT;
   o->rest.i.min = MINIWIN_ANIMATION_TIME_MIN;
   o->rest.i.max = MINIWIN_ANIMATION_TIME_MAX;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_MINDOW_OPACITY];
   o->name = "mindow_opacity";
   o->shortDesc = "Default opacity";
   o->longDesc = "Default opacity for all mindows";
   o->type = CompOptionTypeInt;
   o->value.i = MINIWIN_MINDOW_OPACITY_DEFAULT;
   o->rest.i.min = MINIWIN_MINDOW_OPACITY_MIN;
   o->rest.i.max = MINIWIN_MINDOW_OPACITY_MAX;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_OPACITY_ON_HOVER];
   o->name = "opacity_on_hover";
   o->shortDesc = "Opacity of hovered mindows";
   o->longDesc = "Opacity of hovered mindows";
   o->type = CompOptionTypeInt;
   o->value.i = MINIWIN_OPACITY_ON_HOVER_DEFAULT;
   o->rest.i.min = MINIWIN_OPACITY_ON_HOVER_MIN;
   o->rest.i.max = MINIWIN_OPACITY_ON_HOVER_MAX;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_TASKBAR_SPACING];
   o->name = "taskbar_spacing";
   o->shortDesc = "Space between mindows in the taskbar";
   o->longDesc = "Space between mindows in the taskbar (taskbar mode only)";
   o->type = CompOptionTypeInt;
   o->value.i = MINIWIN_TASKBAR_SPACING_DEFAULT;
   o->rest.i.min = MINIWIN_TASKBAR_SPACING_MIN;
   o->rest.i.max = MINIWIN_TASKBAR_SPACING_MAX;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_DOCK_OPACITY];
   o->name = "dock_opacity";
   o->shortDesc = "Opacity of the dock";
   o->longDesc = "Opacity of the dock (taskbar mode only)";
   o->type = CompOptionTypeInt;
   o->value.i = MINIWIN_DOCK_OPACITY_DEFAULT;
   o->rest.i.min = MINIWIN_DOCK_OPACITY_MIN;
   o->rest.i.max = MINIWIN_DOCK_OPACITY_MAX;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_ONLY_MOVE_UP];
   o->name = "only_move_up";
   o->shortDesc = "Keep mindows on top";
   o->longDesc = "Keep mindows always on top";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_ONLY_MOVE_UP_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_MINDOWS_ALWAYS_ON_TOP];
   o->name = "always_on_top";
   o->shortDesc = "Keep mindows on top";
   o->longDesc = "Keep mindows always on top";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_MINDOWS_ALWAYS_ON_TOP_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_MINDOWS_ON_ALL_WORKSPACES];
   o->name = "on_all_workspaces";
   o->shortDesc = "Mindows on all workspaces";
   o->longDesc = "Mindows on all workspaces";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_MINDOWS_ON_ALL_WORKSPACES_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_USE_AS_TASKBAR];
   o->name = "use_as_taskbar";
   o->shortDesc = "Enable taskbar mode";
   o->longDesc = "Enabel taskbar mode";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_USE_AS_TASKBAR_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_SHOW_DOCK];
   o->name = "show_dock";
   o->shortDesc = "Show the dock window";
   o->longDesc = "Show the dock window (taskbar mode only)";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_SHOW_DOCK_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_DOCK_HIDE_ON_CLICK];
   o->name = "hide_dock_on_click";
   o->shortDesc = "Hide the dock on a left-click";
   o->longDesc = "Hide the dock on a left-click";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_DOCK_HIDE_ON_CLICK_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_RUNS_ON_AIGLX];
   o->name = "runs_on_aiglx";
   o->shortDesc = "Is it AIGLX?";
   o->longDesc = "Is it AIGLX?";
   o->type = CompOptionTypeBool;
   o->value.b = MINIWIN_RUNS_ON_AIGLX_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_MINDOW_KEY];
   o->name = "mindow_key";
   o->shortDesc = "Puts window in mindow mode";
   o->longDesc = "Puts window in mindow mode";
   o->type = CompOptionTypeBinding;
   o->value.bind.type = CompBindingTypeButton;
   o->value.bind.u.button.modifiers = MINIWIN_MINDOW_MODIFIERS_DEFAULT;
   o->value.bind.u.button.button = MINIWIN_MINDOW_BUTTON_DEFAULT;
      
   o = &ms->opt[MINIWIN_SCREEN_OPTION_DEMINDOW_KEY];
   o->name = "demindow_key";
   o->shortDesc = "Restores a mindow's window";
   o->longDesc = "Restores a mindow's window";
   o->type = CompOptionTypeBinding;
   o->value.bind.type = CompBindingTypeButton;
   o->value.bind.u.button.modifiers = MINIWIN_DEMINDOW_MODIFIERS_DEFAULT;
   o->value.bind.u.button.button = MINIWIN_DEMINDOW_BUTTON_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_DOCK_SET_BACKGROUND_KEY];
   o->name = "set_dock_background_key";
   o->shortDesc = "The dock changes its background to a dynamic one";
   o->longDesc = "The dock changes its background to a dynamic one";
   o->type = CompOptionTypeBinding;
   o->value.bind.type = CompBindingTypeButton;
   o->value.bind.u.button.modifiers = MINIWIN_DOCK_SET_BACKGROUND_MODIFIERS_DEFAULT;
   o->value.bind.u.button.button = MINIWIN_DOCK_SET_BACKGROUND_BUTTON_DEFAULT;
   
   o = &ms->opt[MINIWIN_SCREEN_OPTION_WINDOW_TYPE];
   o->name = "window_type";
   o->shortDesc = "Window Types";
   o->longDesc = "Window types that can be iconified (be rather conservative with that)";
   o->type = CompOptionTypeList;
   o->value.list.type = CompOptionTypeString;
   o->value.list.nValue = N_WIN_TYPE;
   o->value.list.value = malloc (sizeof (CompOptionValue) * N_WIN_TYPE);
   for (i = 0; i < N_WIN_TYPE; ++i)
      o->value.list.value[i].s = strdup (winType[i]);
   o->rest.s.string = windowTypeString;
   o->rest.s.nString = nWindowTypeString;
   
   ms->wMask = compWindowTypeMaskFromStringList (&o->value);      
}

static Bool
miniwinInit (CompPlugin* p)
{
   displayPrivateIndex = allocateDisplayPrivateIndex ();
   if (displayPrivateIndex < 0)
   {
      return FALSE;
   }
   return TRUE;
}

static void
miniwinFini (CompPlugin *p)
{
   if (displayPrivateIndex >= 0)
      freeDisplayPrivateIndex (displayPrivateIndex);
}

static Bool
miniwinInitDisplay (CompPlugin* p, CompDisplay* d)
{
   MiniwinDisplay *md;
   
   md = malloc (sizeof (MiniwinDisplay));
   if (!md) return FALSE;
   
   md->screenPrivateIndex = allocateScreenPrivateIndex (d);
   if (md->screenPrivateIndex < 0)
   {
      free (md);
      return FALSE;
   }
   
   WRAP (md, d, handleEvent, miniwinHandleEvent);
   d->privates[displayPrivateIndex].ptr = md;
   
   return TRUE;
}

static void
miniwinFiniDisplay (CompPlugin* p, CompDisplay* d)
{
   MINIWIN_DISPLAY(d);
   
   freeScreenPrivateIndex (d, md->screenPrivateIndex);
   
   UNWRAP (md, d, handleEvent);
   
   free (md);
}



static Bool
miniwinInitScreen (CompPlugin* p, CompScreen* s)
{
   MiniwinScreen* ms;
   MINIWIN_DISPLAY (s->display);
   
   ms = malloc (sizeof (MiniwinScreen));
   if (!ms) return FALSE;
   
   ms->windowPrivateIndex = allocateWindowPrivateIndex (s);
   if (ms->windowPrivateIndex < 0)
   {
      free (ms);
      return FALSE;
   }   
   
   ms->dockWindow = 0;
   
   //General mindow settings
   ms->mindowWidth = MINIWIN_MINDOW_WIDTH_DEFAULT;
   ms->animationTime = MINIWIN_ANIMATION_TIME_DEFAULT;
   ms->mindowOpacity = MINIWIN_MINDOW_OPACITY_DEFAULT;
   ms->scaleOnHover = MINIWIN_SCALE_ON_HOVER_DEFAULT;
   ms->opacityOnHover = MINIWIN_OPACITY_ON_HOVER_DEFAULT;
   ms->mindowsAlwaysOnTop = MINIWIN_MINDOWS_ALWAYS_ON_TOP_DEFAULT;
   ms->mindowsOnAllWorkspaces = MINIWIN_MINDOWS_ON_ALL_WORKSPACES_DEFAULT;
      
   //Taskbar settings
   ms->useAsTaskbar = MINIWIN_USE_AS_TASKBAR_DEFAULT;
   ms->taskbarSpacing = MINIWIN_TASKBAR_SPACING_DEFAULT;
   ms->showDock = MINIWIN_SHOW_DOCK_DEFAULT;
   ms->dockOpacity = MINIWIN_DOCK_OPACITY_DEFAULT;
   ms->dockHideOnClick = MINIWIN_DOCK_HIDE_ON_CLICK_DEFAULT;
   
   //Internal settings
   ms->runsOnAiglx = MINIWIN_RUNS_ON_AIGLX_DEFAULT;
   ms->pendingMindows = 0;
   
   ms->eventQueue = 0;
   ms->dock = 0;   

   ms->mindows = malloc (sizeof(MindowSlot));
   if (!ms->mindows) return FALSE;
   ms->mindows->next = 0;
   ms->mindows->mindow = 0;
   ms->mindows->window = 0;
   
   ms->taskbar = malloc (sizeof(MindowSlot));
   if (!ms->taskbar) return FALSE;
   ms->taskbar->next = 0;
   ms->taskbar->mindow = 0;
   ms->taskbar->window = 0;
   
   ms->nSlotsPerRow = s->width / ( ms->mindowWidth * 2 + ms->taskbarSpacing );
      
   miniwinScreenInitOptions (ms);
   
   s->privates[md->screenPrivateIndex].ptr = ms;

   WRAP (ms, s, paintWindow, miniwinPaintWindow);
   WRAP (ms, s, preparePaintScreen, miniwinPreparePaintScreen);
   WRAP (ms, s, paintScreen, miniwinPaintScreen);
      
   return TRUE;
}

static void
miniwinFiniScreen (CompPlugin* p, CompScreen* s)
{
   MINIWIN_SCREEN(s);
   
   removeAllSlots (ms->taskbar);
   freeWindowPrivateIndex (s, ms->windowPrivateIndex);
   
   while (miniwinPendingEvent(s))
      miniwinPopEvent (s);
   
   UNWRAP (ms, s, paintWindow);
   UNWRAP (ms, s, preparePaintScreen);
   UNWRAP (ms, s, paintScreen);
   
   free (ms);
}

static Bool
miniwinInitWindow (CompPlugin *p, CompWindow *w)
{
   MiniwinWindow *mw;

   MINIWIN_SCREEN (w->screen);

   mw = malloc (sizeof (MiniwinWindow));
   if (!mw) return FALSE;

   mw->state = MINIWIN_WINDOW_STATE_NORMAL_INITIALIZING;      
   mw->hover = FALSE;
   mw->icon = 0;
   mw->parent = 0;
   mw->step = 0;
   
   if (ms->dock == w->id)
   {
      mw->state = MINIWIN_WINDOW_STATE_DOCK_CREATED;
   }
    
   w->privates[ms->windowPrivateIndex].ptr = mw;
   return TRUE;
}

static void
miniwinFiniWindow (CompPlugin *p, CompWindow *w)
{
    MINIWIN_WINDOW (w);
    free (mw);
}

static CompPluginVTable miniwinVTable = {
   "miniwin",
   "Minimizes windows",
   "Minimizes windows and places them at the desktop, similar to gdesklets, but with all ordinary windows",
   miniwinInit,
   miniwinFini,
   miniwinInitDisplay,
   miniwinFiniDisplay,
   miniwinInitScreen,
   miniwinFiniScreen,
   miniwinInitWindow,
   miniwinFiniWindow,
   0, /* GetDisplayOptions */
   0, /* SetDisplayOption */
   miniwinGetScreenOptions,
   miniwinSetScreenOption,
   0, /* deps -- are there any deps? I hope there aren't */
   0, /* ndeps -- is 0 since we have no dependencies... */
};

CompPluginVTable *
getCompPluginInfo (void)
{
   return &miniwinVTable;
}
